#!/usr/bin/env python2
from __future__ import print_function, division
import fcntl, sys#, argparse
import os
import gtk, glib, gobject, pango; from pygtkhelpers import gthreads
import subprocess, dbus; from dbus.mainloop.glib import DBusGMainLoop
import time
import webbrowser
import shelve
import pylast
import re
import urllib
try: import surl
except: pass
try: import keybinder
except: pass
try: import pynotify; pynotify_is_present = True
except: pynotify_is_present = False
#These homeless methods, maybe they'll be adopted by ChipmunkApp
#same goes for icon_map. actually, maybe icon_map will go to GUI

def filter_html_tags(html_str):
    output = html_str
    regex = re.compile('<[^>]*>')
    for substring in regex.findall(html_str):
        output = output.replace(substring, '')
    return output

def extract(source, intro, outro):
    start = source.index(intro) + len(intro)
    end = source.index(outro, start)
    return source[start:end]

def bg(back_task):
    gthreads.AsyncTask(back_task).start()

def play_fetch(master, retriever):
    gthreads.AsyncTask(retriever, master).start()

class Song(object):

    def __init__(self, title='', artist='', album='', duration=0, tags=[], urls={}, lyrics=False):
        self.title = title
        self.artist = artist
        self.album = album
        self.duration = duration
        self.tags = tags
        self.urls = urls
        self.lyrics = lyrics
        if lyrics == False: bg(self.fetch_lyrics)

    def copy(self):
        return Song(self.title, self.artist, self.album, self.duration, self.tags, self.urls, self.lyrics)

    def get_url(self, site_id):
        if site_id == 'youtube': return self.youtube_url
        if site_id == 'beemp3': return self.beemp3_url
        try: return self.urls[site_id]
        except: return None

    def fetch_lyrics(self):
        engines = [self.get_lyrics_lyricstime, self.get_lyrics_lyricsvip]
        for engine in engines:
            try: lyrics, url = engine()
            except: lyrics = ''
            if lyrics != '':
                self.lyrics = lyrics
                self.urls['lyrics'] = url
                break
        else:
            self.lyrics = ''
            query = 'lyrics "{}" "{}"'.format(self.artist, self.title)
            chars = {' ': '+', '&': '%26', '?': '%3F'}
            for bad, good in chars.items(): query = query.replace(bad, good)
            query += '&tbs=prv:1'
            self.urls['lyrics'] = 'http://www.google.com/search?q='+query

    def get_lyrics_lyricsvip(self):
        artist, title = self.artist, self.title
        url = 'http://www.lyricsvip.com/{}/{}-Lyrics.html'.format(artist, title)
        url = url.replace(' ', '-')
        lyrics = extract(urllib.urlopen(url).read(), '</h2>', '&nbsp;')
        lyrics = filter_html_tags(lyrics).strip()
        chars = {'&quot;': '"', '&amp;': '&'}
        for bad, good in chars.items(): lyrics = lyrics.replace(bad, good)
        if lyrics != '': lyrics += '\n\nSource: http://www.lyricsvip.com'
        return lyrics, url

    def get_lyrics_lyricstime(self):
        artist, title = self.artist, self.title
        if artist[-1] == "'": artist = artist[:-1]
        if title[-1] == "'": title = title[:-1]
        url = '{}-{}-lyrics'.format(artist, title).lower()
        for char in ['!', ',', '(', ')', '&', '?']: url = url.replace(char, '')
        for char in ['. ', '.', '  ', ' ', "'"]: url = url.replace(char, '-')
        url = 'http://www.lyricstime.com/{}.html'.format(url)
        lyrics = extract(urllib.urlopen(url).read(), '<div id="songlyrics" >', '</div>')
        lyrics = filter_html_tags(lyrics).strip()
        chars = {'&#039;': "'"}
        for bad, good in chars.items(): lyrics = lyrics.replace(bad, good)
        if lyrics != '': lyrics += '\n\nSource: http://www.lyricstime.com'
        return lyrics, url

#    def get_lyrics_chartlyrics(self):
#        artist, title = self.artist, self.title
#        url = 'http://api.chartlyrics.com/apiv1.asmx/SearchLyricDirect?artist={}&song={}'.format(artist[:75], title[:125])
#        lyrics = extract(urllib.urlopen(url).read(), '<Lyric>', '</Lyric>').strip()
#        if lyrics != '': lyrics += '\n\nSource: http://www.chartlyrics.com'
#        return lyrics, url

    @property
    def lyrics_url(self): return self.get_url('lyrics')

    @property
    def artist_url(self): return self.get_url('artist')

    @property
    def events_url(self): return self.get_url('events')

    @property
    def youtube_url(self):
        query = '"{}" "{}"'.format(self.artist, self.title)
        special_chars = {' ': '+', '&': '%26', '?': '%3F'}
        for bad, good in special_chars.items(): query = query.replace(bad, good)
        return 'http://www.youtube.com/results?search_query='+query

    @property
    def beemp3_url(self):
        query = '"{}" "{}"'.format(self.artist, self.title)
        special_chars = {' ': '+', '&': '%26', '?': '%3F'}
        for bad, good in special_chars.items(): query = query.replace(bad, good)
        return 'http://beemp3.com/index.php?q='+query

icon_map = {'library': 'emblem-sound',
'recommended': 'face-wink',
'neighborhood': 'system-users',
'loved': 'emblem-favorite',
'artist': 'audio-input-microphone',
'tag': 'tomboy',
'group': 'system-users',
'on-the-fly': 'system-users',
'friend': 'user-available',
'neighbor': 'user-available'}

class GUI(object):

    def __init__(self):
        self.fit_on_track_start = False

    def __getitem__(self, item):
        return None

    @property
    def auto_fit(self): return False

    @auto_fit.setter
    def auto_fit(self, indicator_widget): pass

    def clear_progress(self):
        pass

    def fit_window_to_text(self):
        pass

    def hide_menubar(self, bool_or_widget=True):
        return True#if the menubar was hidden

    def set_mini_mode(self, bool_or_widget=True):
        return True#if the user chose mini mode

    def set_track_labels(self, song):
        pass

    def prepare_share_dialog(self):
        pass

class GUI_GTK(GUI, gtk.Builder, object):
    #I'm gonna start migrating gtk-related code into this class

    def __init__(self, markup_file, app):
        super(gtk.Builder, self).__init__()
        self.add_from_file(markup_file)
        self.connect_signals(app)
        self['expander_context'].connect('notify::expanded', self.on_expander_context_notify_expanded)
        self.app = app
        self.auto_fit = False
        self.fit_on_track_start = False

    def __getitem__(self, item):
        return self.get_object(item)

    @property
    def auto_fit(self): return self._auto_fit

    @auto_fit.setter
    def auto_fit(self, widget):
        if type(widget) == type(True):
            self._auto_fit = widget
        else:
            self._auto_fit = widget.get_active()
            tooltip = {True: "Don't Keep Window Fit to Content", False: 'Keep Window Fit to Content'}
            widget.set_tooltip_text(tooltip[self.auto_fit])
        labels = ['label_title', 'label_album', 'label_artist']
        self.ellipsize(labels, not self.auto_fit)
        self['window_main'].set_resizable(not self.auto_fit)

    def clear_progress(self):
        bar = self['progressbar']
        bar.set_fraction(0)
        bar.set_text('')
        bar.set_tooltip_text('Time Remaining')

    def ellipsize(self, label_names, truncate=True):
        mode_map = {True: pango.ELLIPSIZE_END, False: pango.ELLIPSIZE_NONE}
        for label in [self[name] for name in label_names]:
            label.set_ellipsize(mode_map[truncate])
        return False

    def fit_window_to_text(self):
        labels = ['label_title', 'label_album', 'label_artist']
        self.ellipsize(labels, False)
        self.shrink_window()
        glib.idle_add(self.ellipsize, labels)

    def hide_menubar(self, bool_or_widget=True):
        if type(bool_or_widget) != type(True):
            widget = bool_or_widget
            hide = widget.get_active()
        else:
            hide = bool_or_widget
            widget = self['togglebutton_menubar']
        old_text = widget.get_tooltip_text()
        if hide:
            widget.set_tooltip_text(old_text.replace('Hide', 'Show'))
            self['menubar'].hide()
            self.shrink_window('v')
        else:
            widget.set_tooltip_text(old_text.replace('Show', 'Hide'))
            self['menubar'].show()
        return hide

    def set_mini_mode(self, bool_or_widget=True):
        if type(bool_or_widget) == type(True):
            go_tiny = bool_or_widget
        else:
            go_tiny = bool_or_widget.get_active()
        bulky = ['hbox_track_display', 'menubar', 'hbox_status_bar', 'hseparator']
        if go_tiny:
            for w in bulky:
                self[w].hide()
        else:
            for w in bulky:
                self[w].show()
            if self.app.chipmunk.load('hide_menubar'):
                self.hide_menubar()
        self.shrink_window('v')
        return go_tiny

    def set_track_labels(self, song, track_start=True):
        for label_id, data in {'title': song.title, 'artist': song.artist, 'album': song.album}.items():
            data = data.replace('&', '&amp;')
            label = self['label_'+label_id]
            entirety = label.get_label()
            slot = label.get_text().replace('&', '&amp;')
            if slot != '':
                label.set_markup(entirety.replace(slot, data))
            else:
                label.set_markup(entirety.replace('><', '>'+data+'<'))
            label.set_has_tooltip(data != '')
            label.set_tooltip_text(label_id.title()+': '+data.replace('&amp;', '&'))
        if self.fit_on_track_start and track_start:
            self.fit_window_to_text()

    def shrink_window(self, dimension=None):
        widget = self['window_main']
        width, height = widget.get_size()
        if not dimension:
            width, height = 1, 1
        elif dimension == 'v':
            width, height = width, 1
        elif dimension == 'h':
            width, height = 1, height
        widget.resize(width, height)

    def on_expander_context_notify_expanded(self, widget, data):
        expanded = widget.props.expanded
        if not expanded:
            glib.timeout_add(100, self.shrink_window, 'v')#This works MOST of the time...
        self.app.chipmunk.save('show_context', expanded)
#is there a set_track_label_size or something like that?
    def prepare_share_dialog(self):
        self['dialog_share'].set_title('Share '+app.song_to_share.title+' by '+app.song_to_share.artist)

        model = gtk.ListStore(str)
        for friend in self.app.chipmunk.friends: model.append([friend])

        completion = gtk.EntryCompletion()
        completion.set_model(model)
        completion.set_minimum_key_length(1)
        completion.set_text_column(0)
        combo = self['comboboxentry_share']
        combo.set_model(model)
        combo.set_text_column(0)
        combo.child.set_completion(completion)

        self['dialog_share'].show()
        combo.grab_focus()

    def menuitem_station(self, station, label=None):
        if not label: label = station.name
        menu_item = gtk.ImageMenuItem()
        menu_item.set_label(label)
        icon_name = icon_map[station.category]
        image = gtk.image_new_from_icon_name(icon_name, gtk.ICON_SIZE_MENU)
        menu_item.set_image(image)
        menu_item.set_tooltip_text('Play Station')
        menu_item.connect('activate', self.app.on_menuitem_station_clicked, station)
        menu_item.show()
        return menu_item

    def toolbutton_station(self, station, label=None):
        if not label:
            label = station.name
        toolbutton = gtk.ToolButton(None, label)
        toolbutton.set_icon_name(icon_map[station.category])
        toolbutton.set_tooltip_text('Play Station')
        toolbutton.connect('clicked', self.app.on_toolbutton_station_clicked, station)
        toolbutton.set_is_important(True)
        toolbutton.show()
        return toolbutton

class Service(object):
    # Common interface for Lastfm, SomaFM, Pandora, etc.

    @property
    def total_seconds(self): return 0

    @property
    def remaining_seconds(self): return 0

    def play_pause_track(self): pass

    def play_station(self, station): pass

    def stop_track(self): pass

    def ban_track(self): pass

    def login(self, username='', password=''): pass

    def connect(self): pass

    def disconnect(self): pass

    def love_track(self): pass

    def skip_track(self): pass

    @property
    def track(self): return ''

    @property
    def artist(self): return ''

    @property
    def album(self): return ''

    @property
    def artist_url(self): return ''

    @property
    def album_url(self): return ''

    @property
    def track_url(self): return ''

    #@property
    #def station(self): return None

    def get_station(self, type, subject, category): return None

class Lastfm(Service, object):

    def __init__(self, username='<username>', password='<password>',
    host='localhost', port='54311'):
        self.service_id = 'lastfm'
        self.api_key = '04d01b6a9ed86c4eb893dc585ea8d7c4'
        self.api_secret = '9c98b509274bd62a0264f27e50bed381'
        self.username = username
        self.password = password
        self.host = host
        self.port = port
        home_path = os.path.expanduser('~')
        self.chipmunk_path = os.path.join(home_path, '.chipmunk')
        os.putenv('SHELL_FM_HOME', self.chipmunk_path)
        if not os.path.exists(self.chipmunk_path):
            os.makedirs(self.chipmunk_path)
        self.shellfm_config_path = os.path.join(self.chipmunk_path, 'shell-fm.rc')
        self.connected = False
        if username != '<username>':
            self.login(self.username, self.password)
        self.current_song = Song()

    def clobber_shellfm_config(self):
        sep = '[chipmunk-radio-separator]'
        dbus_cmd = 'dbus-send /org/bitbucket/chipmunk org.bitbucket.chipmunk.shellfm_track_started string:'+sep.join(['%t', '%a', '%l', '%d'])
        lines = ['username = '+self.username, 'password = '+self.password,
        #'extern = tee "'+self.chipmunk_path+'/STREAM - %a - %t.mp3" | madplay -Q -',
        #'download = '+self.chipmunk_path+'/%a - %t.mp3',
        #'minimum = 80',
        #'delay-change = something',
        'np-cmd = '+dbus_cmd]
        lines = '\n'.join(lines)
        with open(self.shellfm_config_path, 'w') as f: f.write(lines)

    def login(self, new_username='<username>', new_password='<password>'):
        self.username = new_username
        self.password = new_password
        self.clobber_shellfm_config()
        return self.connect()

    def disconnect(self):
        subprocess.call(['pkill', 'shell-fm'])
        subprocess.call(['pkill', '-9', 'shell-fm'])

    def shellfm_is_running(self):
        output = subprocess.Popen(['pgrep', 'shell-fm'], stdout=subprocess.PIPE).communicate()[0]
        return output != ''

    def connect(self):
        self.disconnect()
        subprocess.call(['shell-fm', '-d', '-i', self.host, '-p', self.port])
        try:
            self.network = self.pylast_network
            self.connected = True
        except:
            self.network = None
            self.connected = False
        return self.connected

    @property
    def pylast_network(self):
        return pylast.get_lastfm_network(api_key=self.api_key,
        api_secret=self.api_secret,
        username=self.username,
        password_hash=pylast.md5(self.password))

    def share_track(self, user_list, message=None, song=None):
        if song == None: song = self.current_song
        track = self.network.get_track(song.artist, song.title)
        if message == '': message = None
        try: track.share(user_list, message)
        except: return False
        #return True

    def shout_track(self, user_list, message='', song=None):
        if song == None: song = self.current_song
        artist, title = song.artist, song.title
        track = '[track artist={}]{}[/track]'.format(artist, title)
        message += '    \n' + track
        for name in user_list:
            if '@' not in name: self.network.get_user(name).shout(message)

    def is_subscriber(self):
        try: return self.network.get_user(self.username).is_subscriber()
        except: return False

    def shellfm(self, command):
        addr = ':'.join(['tcp', self.host, self.port])
        cmd = ['socat', '-t5', '-', addr]
        proc = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        try:
            output, error = proc.communicate(command)
            output = output.strip()
            #print(command, output, error)
        except:
            output, error = '', ''
            #print(command, 'failed')
        else:
            try:
                if output[0] == '"' and output[-1] == '"':
                    output = output[1:-1]#remove surrounding quotes
            except: pass
            if error: print('shell-fm complaint:', error)
        return output

    def play_pause_track(self): self.shellfm('pause')

    def skip_track(self):
        self.shellfm('skip')
        if self.status == 'paused': self.play_pause_track()

    def love_track(self): self.shellfm('love')

    def ban_track(self): self.shellfm('ban')

    def stop_track(self): self.shellfm('stop')

    #def tag_artist(self, tag_list):#works?
        #self.shellfm('tag-artist ' + ','.join(tag_list))

    #def tag_album(self, tag_list):#works?
        #self.shellfm('tag-album ' + ','.join(tag_list))

    def tag_track(self, tag_list):#works?
        self.shellfm('tag-track ' + ','.join(tag_list))

    #def get_artist_tags(self):#worky?
        #return self.shellfm('artist_tags')

    #def get_album_tags(self):#worky?
        #return self.shellfm('album_tags')

    def get_top_tags(self, artist, title, limit):
        try:
            track = self.network.get_track(artist, title)
            top_tag_items = track.get_top_tags(limit)
            return sorted([str(top_item.item) for top_item in top_tag_items], key=str.lower)
        except: return []

    def get_friends(self, limit):
        try:
            friends = self.network.get_user(self.username).get_friends(limit)
            return sorted([str(friend) for friend in friends], key=str.lower)
        except: return []

    #def get_groups(self, limit):
        #try:
            #groups = self.network.get_user(self.username).get_groups(limit)
            #return sorted([str(group) for group in groups], key=str.lower)
        #except: return []

    def get_neighbors(self, limit):
        try:
            neighbors = self.network.get_user(self.username).get_neighbours(limit)
            return sorted([str(neighbor) for neighbor in neighbors], key=str.lower)
        except: return []

    def get_artist_bio(self, artist):
        try:
            bio = self.network.get_artist(artist).get_bio_content()
            return filter_html_tags(bio)
        except: return ''

    def get_artist_events(self, artist):#turns out this is way faster than using pylast calls
        url = 'http://ws.audioscrobbler.com/2.0/?method=artist.getevents&artist={ar}&api_key={k}'
        url = url.format(ar=artist.replace('&', '%26'), k=self.api_key)
        xml_events = urllib.urlopen(url).read().split('</event>')
        #print(xml_events)
        event_dictionaries = []
        for xml in xml_events:
            if '<id>' not in xml: continue
            event = {}
            event['begins'] = extract(xml, '<startDate>', '</startDate>')
            event['title'] = extract(xml, '<title>', '</title>')
            event['description'] = extract(xml, '<description>', '</description>')
            event['description'] = filter_html_tags(event['description']).replace(']]>', '')
            event['headliner'] = extract(xml, '<headliner>', '</headliner>')
            event['url'] = extract(xml, '\n  <url>', '</url>')
            name = extract(xml, '<name>', '</name>')
            city = extract(xml, '<city>', '</city>')
            country = extract(xml, '<country>', '</country>')
            event['venue'] = ', '.join([name, city, country])
            for x in ['title', 'description', 'headliner', 'venue']:
                event[x] = event[x].replace('&amp;', '&').replace('&quot;', '"')
            event_dictionaries.append(event)
        return event_dictionaries

    def get_track_wiki(self, artist, title):
        try:
            wiki = self.network.get_track(artist, title).get_wiki_content()
            return filter_html_tags(wiki)
        except: return ''

    @property
    def cover_image_url(self):
        url = 'http://ws.audioscrobbler.com/2.0/?method=album.getinfo&api_key={}&artist={}&album={}'
        url = url.format(self.api_key, self.current_song.artist.replace('&', '%26'), self.current_song.album.replace('&', '%26'))
        if '#' in url or '(' in url: print(url)
        album_xml = urllib.urlopen(url).read()
        for size in ['extralarge', 'large', 'medium', 'small']:
            try: return extract(album_xml, '<image size="'+size+'">', '</image>')
            except: pass
        return self.shellfm('info "%I"')

    @property
    def artist(self): return self.shellfm('info "%a"')

    @property
    def status(self):
        status = self.shellfm('status').lower()
        if status == '': return 'disconnected'
        return status

    @property
    def volume(self): return self.shellfm('info "%v"')
    @volume.setter
    def volume(self, new_perc): self.shellfm('volume '+str(new_perc * .64))

    @property
    def track(self): return self.shellfm('info "%t"')

    @property
    def album(self): return self.shellfm('info "%l"')

    @property
    def total_seconds(self):
        try: return int(self.shellfm('info "%d"'))
        except: return 0

    @property
    def station_name(self): return self.shellfm('info "%s"')

    @property
    def station_url(self): return self.shellfm('info "%S"')

    @property
    def station(self): self.get_station()

    def get_station(self, station_type=None, subject=None, category=None):

        def urlify(station_type, subject):
            base = 'lastfm://'
            user_type_map = {'library': '/personal',
            'recommended': '/recommended',
            'neighborhood': '/neighbours',
            'loved': '/loved',
            'mix': '/mix'}
            general_type_map = {'artist': 'artist/',
            'tag': 'globaltags/',
            'group': 'group/'}
            if station_type in user_type_map:
                return base + 'user/' + subject + user_type_map[station_type]
            else: return base + general_type_map[station_type] + subject

        if not station_type:
            name = self.station_name
            url = self.station_url
        else:
            if type(subject) == type(''):
                subject = subject.strip().lower()
                formatted = ' '
                for char in subject:
                    if formatted[-1] in [' ', '/', '.']:
                        formatted += char.upper()
                    else: formatted += char
                subject = formatted.strip()
            if type(subject) == type([]):
                category = 'on-the-fly'
                name = ', '.join(subject)
                url = [urlify(station_type, x) for x in subject]
            else:
                if not category: category = station_type
                name = subject
                url = urlify(station_type, subject)
        return Station(self.service_id, name, url, category)

    @property
    def album_url(self): return self.shellfm('info "%L"')

    @property
    def artist_url(self): return self.shellfm('info "%A"')

    @property
    def track_url(self): return self.shellfm('info "%T"')

    @property
    def remaining_seconds(self):
        try: return int(self.shellfm('info "%R"'))
        except: return 0

    def toggle_discovery_mode(self):
        new_state = self.shellfm('discovery')
        if new_state == 'DISCOVERY ON': return True
        elif new_state == 'DISCOVERY OFF': return False

    def toggle_scrobble(self):
        new_state = self.shellfm('rtp')
        if new_state == 'RTP ON': return True
        elif new_state == 'RTP OFF': return False

    #def download_track(self, destination): pass

    def play_station(self, station):
        self.shellfm('play '+station.url)
        if self.status == 'paused': self.play_pause_track()

class Station(object):

    def __init__(self, service_id, name, url, category=None):
        self.service_id = service_id
        self.name = name
        self.category = category
        if type(url) == type([]):
            self.url_list = url
            self.url = self.url_list[0]
        else:
            self.url_list = None
            self.url = url
        self.image = None

    def rotate(self):
        if self.url_list:
            self.url_list = self.url_list[1:] + self.url_list[:1]
            self.url = self.url_list[0]
            return True
        return False

    def __eq__(self, other):
        try:
            if other == None: return False
        except: pass
        if self.category == 'on-the-fly' and other.category == 'on-the-fly':
            if set(self.url_list) == set(other.url_list): return True
        properties = [(self.service_id, other.service_id),
        (self.name.lower(), other.name.lower()),
        (self.category, other.category)]
        for a, b in properties:
            if a != b: return False
        return True

    def __ne__(self, other): return not self.__eq__(other)

    def __str__(self): return self.name


class Chipmunk(object):

    def __init__(self):
        self.config_dir = os.path.join(os.path.expanduser('~'), '.chipmunk')
        if not os.path.isdir(self.config_dir): os.mkdir(self.config_dir)
        self.db = os.path.join(self.config_dir, 'chipmunk_data')
        self.services = {}
        self.services['lastfm'] = Lastfm()
        self.current_station = None
        self._current_song = Song()
        #self.cover_image_url = ''
        self.queued_task = None
        self.time = 0
        self.friends = []

    #def refresh(self):
        #subprocess.call(['dbus-send', '/org/bitbucket/chipmunk',
        #'org.bitbucket.chipmunk.shellfm_track_started'])

    @property
    def status(self): return self.services['lastfm'].status

    def get_status(self): return self.status

    @property
    def station(self): return self.get_station()

    def get_station(self, station_type=None, subject=None, category=None, service_id='lastfm'):
        return self.services[service_id].get_station(station_type, subject, category)

    def is_subscriber(self):
        service = self.services['lastfm']
        return service.is_subscriber()

    def toggle_discovery_mode(self, service_id='lastfm'):
        if service_id == 'lastfm':
            mode = self.services[service_id].toggle_discovery_mode()
            return mode

    def toggle_scrobble(self, service_id='lastfm'):
        if service_id == 'lastfm':
            return self.services[service_id].toggle_scrobble()

    def save(self, key, value):
        shelf = shelve.open(self.db)
        shelf[key] = value
        shelf.close()

    @property
    def username(self): return self.get_username()

    def get_username(self, service_id='lastfm'):
        return self.services[service_id].username

    @property
    def password(self): return self.get_password()

    def get_password(self, service_id='lastfm'):
        return self.services[service_id].password

    def load(self, key):
        shelf = shelve.open(self.db)
        if key == 'auto_fit': return False
        try: value = shelf[key]
        except:#I should probably hide all load/save calls behind properties, putting defaults there.
        #saved_recent_artists, saved_auto_fit, etc.
            default_map = {
            'lastfm_username': '<username>',
            'lastfm_password': '<password>',
            'auto_fit': False,
            'hide_menubar': False,
            'show_context': False,
            'discovery': False,
            'cover_size': 96,
            'recent_stations': [],
            'recent_artists': [],
            'recent_tags': [],
            'recent_groups': [],
            'window_size': (775, 1),
            'bookmarks': [],
            'shout_when_share': True,
            'is_subscriber': False,
            'label_track_size': 22,
            'fit_on_track_start': False,
            'notifications_enabled': True,
            'mini': False}
            try: value = default_map[key]
            except: value = None
        shelf.close()
        return value

    @property
    def recent_stations(self): return self.load('recent_stations')
    @recent_stations.setter
    def recent_stations(self, stations): self.save('recent_stations', stations)

    @property
    def recent_artists(self): return self.load('recent_artists')
    @recent_artists.setter
    def recent_artists(self, artists): self.save('recent_artists', artists)

    @property
    def recent_tags(self): return self.load('recent_tags')
    @recent_tags.setter
    def recent_tags(self, tags): self.save('recent_tags', tags)

    @property
    def recent_groups(self): return self.load('recent_groups')
    @recent_groups.setter
    def recent_groups(self, groups): self.save('recent_groups', groups)

    def play_pause_track(self): self.service.play_pause_track()

    def notify(self, message=None, sub=None, icon='chipmunk-radio', timeout=5000):
        if self.load('notifications_enabled') and pynotify_is_present:
            if message == None:
                title = self.current_song.title
                body = self.current_song.artist
            else: title, body = message, sub
            try:
                notification = pynotify.Notification(title, body, icon)
                notification.set_timeout(timeout)
                notification.show()
            except: pass

    @property
    def remaining_seconds(self): return self.service.remaining_seconds

    @property
    def total_seconds(self): return self.service.total_seconds

    def play_station(self, station):
        self.current_station = station
        self.services[station.service_id].play_station(station)

    def stop_track(self): self.service.stop_track()

    def ban_track(self):
        self.notify('Track Banned', self.current_song.title, 'emblem-unreadable')
        self.service.ban_track()

    def set_top_tags(self):
        song = self.current_song
        song.tags = self.top_tags
        self.current_song = song

    def share_track(self, user_list, message=None, song=None):
        self.service.share_track(user_list, message, song)

    def shout_track(self, user_list, message='', song=None):
        self.service.shout_track(user_list, message, song)

    def login(self, service_id='lastfm', new_username=None, new_password=None):
        if not new_username: new_username = self.load(service_id+'_username')
        if not new_password: new_password = self.load(service_id+'_password')
        self.save(service_id+'_username', new_username)
        self.save(service_id+'_password', new_password)
        service = self.services[service_id]
        service.login(new_username, new_password)
        if service_id == 'lastfm':
            self.current_station = self.get_station('mix', service.username)

    def disconnect(self, service_id='lastfm'):
        self.services[service_id].disconnect()

    @property
    def cover_image(self):
        url = self.cover_image_url
        if not url: return None
        filepath = os.path.join(self.config_dir, 'cover.jpg')
        image = urllib.URLopener()#aria2c?
        try:
            image.retrieve(url, filepath)#aria2c?
            return filepath
        except: return None

    def get_cover_image(self): return self.cover_image

    @property
    def cover_image_url(self): return self.service.cover_image_url

    def get_cover_image_url(self): return self.cover_image_url

    def clear_queue(self): self.queued_task = None

    @property
    def current_song(self): return self._current_song# I really gotta stop duplicating the current_song... especially now that each song fetches its own lyrics

    @current_song.setter
    def current_song(self, song):
        self._current_song = song.copy()
        self.service.current_song = song.copy()

    def process_queue(self):
        if self.queued_task == None: return False
        task = self.queued_task
        self.clear_queue()
        task[0](*task[1])
        if task[0] == self.stop_track:
            self.notify('Stopped', '', 'media-playback-stop')
        return True

    def queue_station(self, station):
        self.queued_task = (self.play_station, [station])

    def queue_stop(self): self.queued_task = (self.stop_track, [])

    def love_track(self):
        self.service.love_track()
        self.notify('Track Loved', self.current_song.title, 'emblem-favorite')

    def skip_track(self): self.service.skip_track()

    @property
    def track(self): return self.service.track

    @property
    def top_tags(self): return self.get_top_tags()

    def get_top_tags(self, limit=3):
        try: return self.service.get_top_tags(self.current_song.artist, self.current_song.title, limit)
        except: return []

    @property
    def top_tag(self):
        try: return self.get_top_tags(1)[0]
        except: return None

    def get_friends(self, limit=100):
        try: self.friends = self.service.get_friends(limit)
        except: self.friends = []
        return self.friends

    @property
    def neighbors(self): return self.get_neighbors()

    def get_neighbors(self, limit=50):
        try: return self.service.get_neighbors(limit)
        except: return []

    @property
    def artist(self): return self.service.artist

    @property
    def album(self): return self.service.album

    @property
    def album_url(self): return self.service.album_url

    @property
    def artist_url(self): return self.service.artist_url

    @property
    def volume(self): return self.service.volume

    @volume.setter
    def volume(self, vol): self.service.volume = vol

    @property
    def track_url(self): return self.service.track_url

    @property
    def service(self): return self.services[self.current_station.service_id]

    def get_lyrics(self):
        while self.current_song.lyrics == False: time.sleep(1)
        return self.current_song.lyrics

    @property
    def artist_bio(self): return self.get_artist_bio()

    def get_artist_bio(self):
        return self.service.get_artist_bio(self.current_song.artist)

    @property
    def artist_events(self):
        return self.service.get_artist_events(self.current_song.artist)

    def get_artist_events(self): return self.artist_events

    @property
    def track_wiki(self): return self.get_track_wiki()

    def get_track_wiki(self):
        return self.service.get_track_wiki(self.current_song.artist, self.current_song.title)

    def bookmark_station(self, station):
        bookmarks = self.load('bookmarks')
        if station not in bookmarks:
            bookmarks.append(station)
            self.save('bookmarks', bookmarks)
        self.notify('Station Bookmarked', station.name, 'bookmark-new')

    def debookmark_station(self, station):
        bookmarks = self.load('bookmarks')
        try: bookmarks.remove(station)
        except: pass
        self.save('bookmarks', bookmarks)
        self.notify('Station Removed from Bookmarks', station.name, 'list-remove')


class ChipmunkApp(object):

    def __init__(self):
        self.still_initting = True
        self.auto_play = False
        self.stopped = True
        self.last_volume, self.target_volume = 100, 100
        self.gui = GUI_GTK('chipmunk-radio.glade', self)
        self.chipmunk = Chipmunk()
        self.load_settings()
        self.init_daemons()

    def init_daemons(self):
        pynotify.init('Chipmunk Radio')
        DBusGMainLoop(set_as_default=True)
        self.session_bus = dbus.SessionBus()
        self.session_bus.add_signal_receiver(self.on_track_start, signal_name='shellfm_track_started')
        self.bind_keys()
        bg(self.update_progress_loop)

    def build_menu(self, menuitem_name, category, recents, context_stations, new_handlers):
    #destined for GUI_GTK, with others...
        def add_sep(menu, do_it=True):
            if do_it:
                menu_item = gtk.SeparatorMenuItem()
                menu.append(menu_item)
                menu_item.show()

        bookmarks = sorted(self.chipmunk.load('bookmarks'), key=str)
        parent = self.gui[menuitem_name]
        new_menu = gtk.Menu()

        need_separator = False
        for station in context_stations:
            #if station not in bookmarks + recents + [self.chipmunk.current_station]:
            new_menu.append(self.gui.menuitem_station(station))
            need_separator = True
        add_sep(new_menu, need_separator)

        need_separator = False
        for station in bookmarks:
            cat = station.category
            if cat == 'on-the-fly': cat = 'group'
            if cat == category:
                new_menu.append(self.gui.menuitem_station(station))
                need_separator = True
        add_sep(new_menu, need_separator)

        need_separator = False
        recents_rvrsd = recents[:]; recents_rvrsd.reverse()
        for station in recents_rvrsd:
            if station not in bookmarks + [self.chipmunk.current_station]:
                new_menu.append(self.gui.menuitem_station(station))
                need_separator = True
        add_sep(new_menu, need_separator)

        for text, handler in new_handlers:
            menu_item = gtk.ImageMenuItem(text)
            new_image = gtk.image_new_from_icon_name('list-add', gtk.ICON_SIZE_MENU)
            menu_item.set_image(new_image)
            menu_item.connect('activate', handler)
            new_menu.append(menu_item)
            menu_item.show()

        parent.deselect()
        parent.remove_submenu()
        parent.set_submenu(new_menu)

#todo: reduce number of 'app' accesses in GUI_GTK, preferring to pass in necessary info
    def build_artists_menu(self):
        menuitem_name = 'menuitem_artists'
        category = 'artist'
        recents = self.chipmunk.load('recent_artists')
        context_stations = []
        artist = self.chipmunk.current_song.artist.strip()
        if artist != '':
            context_stations.append(self.chipmunk.get_station('artist', artist))
        new_handlers = [('_New Artist', self.on_menuitem_artists_new_activate)]

        self.build_menu(menuitem_name, category, recents, context_stations, new_handlers)

    def build_tags_menu(self, top_tags):
        menuitem_name = 'menuitem_tags'
        category = 'tag'
        recents = self.chipmunk.load('recent_tags')
        context_stations = []
        for tag in top_tags:
            context_stations.append(self.chipmunk.get_station('tag', tag))
        new_handlers = [('_New Tag', self.on_menuitem_tags_new_activate)]

        self.build_menu(menuitem_name, category, recents, context_stations, new_handlers)

    def build_groups_menu(self, group_stations=[]):
        menuitem_name = 'menuitem_groups'
        category = 'group'
        recents = self.chipmunk.load('recent_groups')
        context_stations = group_stations
        new_handlers = [('_New Group', self.on_menuitem_groups_new_activate)]
        new_handlers.append(('New On-the-_Fly Group', self.on_menuitem_groups_rotation_new_activate))

        self.build_menu(menuitem_name, category, recents, context_stations, new_handlers)

    def update_top_tag_toolbutton(self):
        try: self.build_top_tag_toolbutton(self.chipmunk.current_song.tags[0])
        except: self.build_top_tag_toolbutton(None)

    def build_top_tag_toolbutton(self, tag=None):
        widget = self.gui['toolbutton_top_tag']
        widget.hide()
        if tag != None:
            station = self.chipmunk.get_station('tag', tag)
            if self.chipmunk.current_station != station:
                widget.set_label(station.name)
                widget.show()

    def update_all_menus(self):
        self.update_artists_menu()
        self.update_tags_menu()
        self.update_groups_menu()
        self.update_friends_menu()
        self.update_neighbors_menu()

    def update_menu(self, category):
        menu_map = {'friend': self.update_friends_menu,
        'tag': self.update_tags_menu, 'artist': self.update_artists_menu,
        'group': self.update_groups_menu,
        'on-the-fly': self.update_groups_menu,
        'neighbor': self.update_neighbors_menu}
        menu_map[category]()

    def load_offline_settings(self):#this (or most of this) will likely move into GUI_GTK soon
        self.gui['togglebutton_menubar'].set_active(self.chipmunk.load('hide_menubar'))
        self.gui['expander_context'].set_expanded(self.chipmunk.load('show_context'))
        self.cover_size = self.chipmunk.load('cover_size')
        self.gui['image_cover'].set_pixel_size(self.cover_size)
        self.label_track_size = self.chipmunk.load('label_track_size')
        self.on_vbox_track_display_scroll_event(None, None)
        self.gui['window_main'].resize(*self.chipmunk.load('window_size'))
        self.gui['togglebutton_mini'].set_active(self.chipmunk.load('mini'))
        self.gui['togglebutton_auto_fit'].set_active(self.chipmunk.load('auto_fit'))
        self.update_recent_toolbar()
        self.update_artists_menu()
        self.update_groups_menu()
        self.update_tags_menu()
        subscribes = self.chipmunk.load('is_subscriber')
        self.gui['menuitem_my_loved'].set_sensitive(subscribes)
        self.gui['menuitem_guest_loved'].set_sensitive(subscribes)
        self.gui['menuitem_notifications'].set_active(self.chipmunk.load('notifications_enabled'))
        self.gui['menuitem_fit'].set_active(self.chipmunk.load('fit_on_track_start'))
        self.gui['checkbutton_share_shout'].set_active(self.chipmunk.load('shout_when_share'))

    def load_online_settings(self):
        self.set_subscriber_settings()
        self.update_friends_menu()
        self.update_neighbors_menu()
        self.still_initting = False
        if self.auto_play:
            self.on_button_play_pause_clicked()
        self.auto_play = False

    def connect(self, username=None, password=None, service_id='lastfm'):
        if username == None:
            username = self.chipmunk.load('lastfm_username')
            password = self.chipmunk.load('lastfm_password')
        self.chipmunk.login(service_id, username, password)
        self.load_online_settings()

    def load_settings(self):
        self.load_offline_settings()
        bg(self.connect)

    def on_dbus_mm_key(self, *data):
        bindings = {'Play': self.on_button_play_pause_clicked,
        'Stop': self.toggle_stop_delay,
        'Next': self.on_button_skip_clicked,
        'Previous': self.chipmunk.notify,
        'Share': self.on_button_share_clicked,
        'Love': self.on_button_love_clicked,
        'Ban': self.on_button_ban_clicked}
        for key in data:
            print(key)
            try: bindings[key]()
            except: pass

    def on_window_main_key_event(self, window, data, depressed=True):
        key = extract(str(data), 'keyval=', '>')
        if key in ['Control_L', 'Control_R']: self.key_ctrl = depressed
        elif key in ['N', 'n']: self.key_n = depressed

    def on_window_main_key_press_event(self, window, data):
        self.on_window_main_key_event(window, data, True)
        try:
            if self.key_ctrl and self.key_n:
                self.on_toolbutton_new_station_inline_clicked(None)
        except: pass

    def on_window_main_key_release_event(self, window, data):
        self.on_window_main_key_event(window, data, False)

    def bind_keys(self):
        try: keybinder.bind('XF86AudioPlay', self.on_button_play_pause_clicked)
        except: pass
        try: keybinder.bind('XF86AudioStop', self.toggle_stop_delay)
        except: pass
        try: keybinder.bind('XF86AudioNext', self.on_button_skip_clicked)
        except: pass
        try: keybinder.bind('XF86AudioPrev', self.chipmunk.notify)
        except: pass
        self.session_bus.add_signal_receiver(self.on_dbus_mm_key, signal_name='MediaPlayerKeyPressed')

    def toggle_stop_delay(self):
        widget = self.gui['togglebutton_stop_delay']
        if widget.props.sensitive: widget.set_active(not widget.get_active())

    def bookmark_station(self):
        station = self.chipmunk.current_station
        self.chipmunk.bookmark_station(station)
        self.update_menu(station.category)
        self.gui['toolbutton_bookmark'].hide()
        self.gui['toolbutton_debookmark'].show()

    def debookmark_station(self):
        station = self.chipmunk.current_station
        self.chipmunk.debookmark_station(station)
        self.update_menu(station.category)
        self.gui['toolbutton_debookmark'].hide()
        self.gui['toolbutton_bookmark'].show()

    def add_recent(self, station, category=None, max=20):
        cat_map = {
        'artist': ('recent_artists', self.update_artists_menu),
        'tag': ('recent_tags', self.update_tags_menu),
        'group': ('recent_groups', self.update_groups_menu),
        'on-the-fly': ('recent_groups', self.update_groups_menu),
        None: ('recent_stations', None)}
        if category in cat_map:
            recent_name, recent_update = cat_map[category]
            recent_list = self.chipmunk.load(recent_name)
            try: recent_list.remove(station)
            except: pass
            recent_list.append(station)
            if len(recent_list) > max: recent_list = recent_list[1:]
            self.chipmunk.save(recent_name, recent_list)
            if recent_update != None: recent_update()
            if not category: self.add_recent(station, station.category)

    def resize(self, image_widget, new_size, image_file=None):
        image_widget.set_pixel_size(new_size)
        if image_file:
            pixbuf = gtk.gdk.pixbuf_new_from_file_at_size(image_file, new_size, new_size)
            image_widget.set_from_pixbuf(pixbuf)
        self.cover_size = new_size


    def on_vbox_track_display_scroll_event(self, widget, event):#I should probably implement a "lock" on this.
        step = 1
        if event == None: step = 0
        elif event.direction in [gtk.gdk.SCROLL_DOWN, gtk.gdk.SCROLL_LEFT]:
            step = -step
        new_size = self.label_track_size + step
        label_info = {'title': ('<span weight="bold" font="'+str(new_size)+'">', '</span>'),
        'artist': ('<span style="italic" font="'+str(new_size)+'">', '</span>'),
        'album': ('<span font="'+str(new_size)+'">', '</span>')}
        for name, markup in label_info.items():
            label = self.gui['label_'+name]
            label.set_markup(label.get_text().join(markup).replace('&', '&amp;'))
        self.label_track_size = new_size

    def on_button_cover_scroll_event(self, widget, event):
        image = self.gui['image_cover']
        old_size = image.get_pixel_size()
        new_size = old_size
        min_size, max_size = 64, 352
        size_step = 32
        if event.direction == gtk.gdk.SCROLL_UP:
            if old_size < max_size: new_size += size_step
        elif event.direction == gtk.gdk.SCROLL_DOWN:
            if old_size > min_size: new_size -= size_step
        if image.get_icon_name()[0]: image_file = None
        else: image_file = os.path.join(self.chipmunk.config_dir, 'cover.jpg')
        self.resize(image, new_size, image_file)

    def on_button_new_station_cancel_clicked(self, widget):
        self.gui['dialog_new_station'].hide()

    def on_button_new_station_play_clicked(self, widget):
        if self.next_station_type == 'rotation':
            stn_type = 'mix'
            widget = self.gui['entry_new_station']
            subject = [user.strip() for user in widget.get_text().split(',') if user.strip() != '']
        else:
            stn_type = self.next_station_type
            subject = self.gui['entry_new_station'].get_text()
        self.next_station = self.chipmunk.get_station(stn_type, subject)
        self.gui['dialog_new_station'].hide()
        self.play_station(self.next_station)

    def on_menuitem_artists_new_activate(self, widget):
        self.new_station('lastfm', 'artist')

    def new_station(self, service_id, stn_type):
        #types: library, recommended, neighborhood, loved, artist, tag, group, mix
        self.next_station_service_id = service_id
        self.next_station_type = stn_type
        label_map = {}
        label_tip_map = {}
        for t in ['library', 'recommended', 'neighborhood', 'loved', 'mix']:
            label_map[t] = 'Username:  '
        label_map['artist'] = 'Artist:  '
        label_map['rotation'] = 'Usernames:  '
        label_tip_map['rotation'] = 'e.g. "EvilGnomeAndy, EvilGnomeCandy, Rockijosh" (without quotes)'
        label_map['group'] = 'Group:  '
        label_map['tag'] = 'Tag:  '
        self.gui['label_new_station'].set_text(label_map[stn_type])
        widget = self.gui['label_new_station_tip']
        try:
            widget.set_text(label_tip_map[stn_type])
            widget.show()
        except: widget.hide()
        self.gui['entry_new_station'].grab_focus()
        self.gui['dialog_new_station'].show()

    def on_button_fit_clicked(self, widget):
        self.gui.fit_window_to_text()

    def on_togglebutton_auto_fit_toggled(self, widget):
        self.gui.auto_fit = widget

#THESE ARE OBVIOUSLY REPETITIVE - improve code!
    def on_menuitem_my_recommendations_activate(self, widget):
        station = self.chipmunk.get_station('recommended', self.chipmunk.username)
        self.play_station(station)

    def on_menuitem_my_neighborhood_activate(self, widget):
        station = self.chipmunk.get_station('neighborhood', self.chipmunk.username)
        self.play_station(station)

    def on_menuitem_my_loved_activate(self, widget):
        station = self.chipmunk.get_station('loved', self.chipmunk.username)
        self.play_station(station)

    def on_menuitem_guest_library_activate(self, widget):
        self.new_station('lastfm', 'library')

    def on_menuitem_guest_mix_activate(self, widget):
        self.new_station('lastfm', 'mix')

    def on_menuitem_guest_recommendations_activate(self, widget):
        self.new_station('lastfm', 'recommended')

    def on_menuitem_guest_neighborhood_activate(self, widget):
        self.new_station('lastfm', 'neighborhood')

    def on_menuitem_guest_loved_activate(self, widget):
        self.new_station('lastfm', 'loved')

    def on_menuitem_tags_new_activate(self, widget):
        self.new_station('lastfm', 'tag')

    def on_menuitem_groups_new_activate(self, widget):
        self.new_station('lastfm', 'group')

    def on_menuitem_groups_rotation_new_activate(self, widget):
        self.new_station('lastfm', 'rotation')

    def gtk_widget_hide_on_delete(self, widget, event):
        widget.hide()
        return True

    def update_progress_loop(self):
        bar = self.gui['progressbar']
        play_pause_img = self.gui['image_play_pause']
        status = self.chipmunk.status#eek! so many shell-fm calls!
        if status == 'playing':
            remaining_seconds = self.chipmunk.remaining_seconds
            if remaining_seconds <= self.chipmunk.current_song.duration and remaining_seconds > 0:
                bar.set_fraction(remaining_seconds/self.chipmunk.current_song.duration)
                bar.set_text(str(remaining_seconds // 60) + ':' + str(remaining_seconds % 60).zfill(2))
            else:
                bar.set_fraction(0)
                bar.set_text('')
            if play_pause_img.get_icon_name()[0] == 'media-playback-start':
                self.button_play_pause_toggle()
        else:
            if play_pause_img.get_icon_name()[0] == 'media-playback-pause':
                self.button_play_pause_toggle()
            if status == 'stopped' and not self.stopped:
                self.on_track_stop()
            elif status != 'paused':
                bar.set_text(status.title())
        #print(self.last_volume, 'vs.', self.target_volume)
        if self.target_volume != self.last_volume:
            self.chipmunk.volume = self.target_volume
            self.last_volume = self.target_volume
        time.sleep(1)
        bg(self.update_progress_loop)

    def on_window_main_destroy(self, widget):
        self.chipmunk.save('window_size', widget.get_size())
        self.chipmunk.save('label_track_size', self.label_track_size)
        self.chipmunk.save('cover_size', self.cover_size)
        self.chipmunk.save('auto_fit', self.gui.auto_fit)
        self.chipmunk.save('fit_on_track_start', self.gui.fit_on_track_start)
        self.chipmunk.disconnect()
        gtk.main_quit()

    def on_button_play_pause_clicked(self, widget=None):
        if self.still_initting: self.auto_play = not self.auto_play
        else:
            status = self.chipmunk.status
            if status in ['playing', 'paused']:
                self.chipmunk.play_pause_track()
            else:
                self.play_station(self.chipmunk.current_station, status)

    def query_delay_station(self, station):
        self.gui['messagedialog_delay'].show()

    def play_station(self, station, status=None):

        def helper(status=None):
            if status == None: self.chipmunk.play_station(self.next_station)
            else:
                if status == 'disconnected': self.login(self.next_station.service_id)
                elif status == 'playing': self.query_delay_station(self.next_station)
                else: bg(helper)

        self.next_station = station
        if status != None: helper(status)
        else: play_fetch(helper, self.chipmunk.get_status)

    def on_menuitem_station_clicked(self, widget, station):
        self.play_station(station)

    def button_play_pause_toggle(self):
        toggle = {'media-playback-start': 'media-playback-pause',
        'media-playback-pause': 'media-playback-start',
        'Play': 'Pause', 'Pause': 'Play'}
        image = self.gui['image_play_pause']
        name, size = image.get_icon_name()
        image.set_from_icon_name(toggle[name], size)
        button = self.gui['button_play_pause']
        button.set_tooltip_text(toggle[button.get_tooltip_text()])

    def on_messagedialog_delay_response(self, widget, response):
        NO, YES, X = -9, -8, -4#there must be a more semantic way...
        if response == YES:
            self.clear_queue()
            self.chipmunk.queue_station(self.next_station)
            self.gui['image_queued'].show()
            self.gui['label_queued'].set_text(self.next_station.name)
            self.gui['button_clear_queue'].show()
            self.chipmunk.notify('Switch After Track',
            'Chipmunk will switch stations when this track ends', 'appointment-new')
        elif response == NO:
            self.clear_queue()
            if self.chipmunk.status == 'playing':
                self.chipmunk.play_pause_track()
            self.play_station(self.next_station)
        widget.hide()

    def on_togglebutton_scrobble_toggled(self, widget):
        if type(self.chipmunk.services['lastfm'].network) != type(None):
            old_text = widget.get_tooltip_text()
            if widget.get_active():
                widget.set_tooltip_text(old_text.replace('On', 'Off'))
                #turn scrobbling on
                if not self.chipmunk.toggle_scrobble():
                    self.chipmunk.toggle_scrobble()
            else:
                widget.set_tooltip_text(old_text.replace('Off', 'On'))
                #turn scrobbling off
                if self.chipmunk.toggle_scrobble():
                    self.chipmunk.toggle_scrobble()
        #else:
            #widget.set_active(not widget.get_active())

    def on_togglebutton_discovery_toggled(self, widget, stop_recursion=False):
        if type(self.chipmunk.services['lastfm'].network) != type(None):
            old_text = widget.get_tooltip_text()
            if widget.get_active():
                widget.set_tooltip_text(old_text.replace('On', 'Off'))
                #turn discovery on
                if not self.chipmunk.toggle_discovery_mode():
                    self.chipmunk.toggle_discovery_mode()
            else:
                widget.set_tooltip_text(old_text.replace('Off', 'On'))
                #turn discovery off
                if self.chipmunk.toggle_discovery_mode():
                    self.chipmunk.toggle_discovery_mode()
        #else:
            #widget.set_active(not widget.get_active())

    def on_togglebutton_mini_toggled(self, widget):
        go_tiny = self.gui.set_mini_mode(widget)
        self.chipmunk.save('mini', go_tiny)

    def on_togglebutton_menubar_toggled(self, widget):
        hidden = self.gui.hide_menubar(widget)
        self.chipmunk.save('hide_menubar', hidden)

    def on_toolbutton_new_station_inline_clicked(self, widget):
        self.next_station_service_id = 'lastfm'
        for w in ['toolbar_recent_stations', 'toolbar_context']:
            self.gui[w].hide()
        self.gui['toolbutton_new_station_inline'].hide()
        entry = self.gui['comboboxentry_new_station_inline'].child
        entry.set_text(self.gui['entry_new_station'].get_text())
        for w in ['comboboxentry_new_station_inline',
        'toolbutton_new_station_inline_cancel', 'toolbutton_new_group',
        'toolbutton_new_artist', 'toolbutton_new_otf_group',
        'toolbutton_new_guest_library', 'toolbutton_new_tag']:
            self.gui[w].show()
        entry.grab_focus()

        #completion:
        model = gtk.ListStore(str)
        for friend in self.chipmunk.friends: model.append([friend])
        for tag in self.chipmunk.recent_tags: model.append([tag])
        for tag in self.chipmunk.current_song.tags: model.append([tag])
        for artist in self.chipmunk.recent_artists: model.append([artist])
        for group in self.chipmunk.recent_groups: model.append([group])
        for station in self.chipmunk.load('bookmarks'): model.append([station.name])
        self.gui['comboboxentry_new_station_inline'].set_model(model)
        self.gui['comboboxentry_new_station_inline'].set_text_column(0)
        completion = gtk.EntryCompletion()
        completion.set_model(model)
        completion.set_minimum_key_length(1)
        completion.set_text_column(0)
        entry.set_completion(completion)

    def on_toolbutton_new_station_inline_cancel_clicked(self, widget):
        for w in ['toolbar_context', 'toolbar_recent_stations',
        'toolbutton_new_station_inline']:
            self.gui[w].show()
        for w in ['comboboxentry_new_station_inline',
        'toolbutton_new_station_inline_cancel', 'toolbutton_new_group',
        'toolbutton_new_artist', 'toolbutton_new_otf_group',
        'toolbutton_new_guest_library', 'toolbutton_new_tag']:
            self.gui[w].hide()

    def on_toolbutton_new_tag_clicked(self, widget):
        self.on_toolbutton_new_clicked('tag')

    def on_toolbutton_new_clicked(self, next_type):
        self.on_toolbutton_new_station_inline_cancel_clicked(None)
        self.next_station_type = next_type
        subject = self.gui['comboboxentry_new_station_inline'].child.get_text()
        self.gui['entry_new_station'].set_text(subject)
        self.on_button_new_station_play_clicked(None)

    def on_toolbutton_new_artist_clicked(self, widget):
        self.on_toolbutton_new_clicked('artist')

    def on_toolbutton_new_otf_group_clicked(self, widget):
        self.on_toolbutton_new_clicked('rotation')

    def on_toolbutton_new_guest_library_clicked(self, widget):
        self.on_toolbutton_new_clicked('library')

    def on_toolbutton_new_group_clicked(self, widget):
        self.on_toolbutton_new_clicked('group')

    def clear_queue(self):
        self.chipmunk.clear_queue()
        self.gui['image_queued'].hide()
        self.gui['label_queued'].set_text('')
        self.gui['button_clear_queue'].hide()
        self.gui['togglebutton_stop_delay'].set_active(False)

    def on_togglebutton_stop_delay_toggled(self, widget):
        old_text = widget.get_tooltip_text()
        if widget.get_active():
            self.chipmunk.queue_stop()
            widget.set_tooltip_text("Don't " + old_text)
            self.gui['image_queued'].show()
            self.gui['label_queued'].set_text('Stop')
            self.gui['button_clear_queue'].show()
            self.chipmunk.notify('Stop After Track',
            'Chipmunk will stop when this track ends', 'appointment-new')
        else:
            self.clear_queue()
            widget.set_tooltip_text(old_text.replace("Don't ", ''))

    def set_cover_image_from_file(self, image_file=None):
        widget = self.gui['image_cover']
        if image_file == None:
            try: station_image = self.chipmunk.current_station.image
            except: station_image = None
            if not station_image:
                widget.set_from_icon_name('media-optical', self.cover_size)
                self.tooltip_cover = None
                self.gui['label_expander_context'].set_has_tooltip(False)
                self.gui['button_cover'].set_has_tooltip(False)
            else: pass
        else:
            self.tooltip_cover = gtk.gdk.pixbuf_new_from_file_at_size(image_file, 500, 500)
            try:
                self.gui['label_expander_context'].set_has_tooltip(True)
                self.gui['button_cover'].set_has_tooltip(True)
            except: pass
        self.resize(widget, self.cover_size, image_file)

    def on_button_cover_query_tooltip(self, widget, x, y, kb_mode, tooltip):
        if not widget.props.sensitive: return False
        #the text is annoying next to the pic, I may have to make a custom window
        #to put it below...
        #tooltip.set_text("View Album on Last.fm\nScroll to Resize")
        try: tooltip.set_icon(self.tooltip_cover)
        except: tooltip.set_icon(None)
        return True

    def on_label_expander_context_query_tooltip(self, widget, x, y, kb_mode, tooltip):
        return self.on_button_cover_query_tooltip(self.gui['button_cover'], x, y, kb_mode, tooltip)

    def set_cover_image(self, get_url=False):
        #print('setting cover image: getting url:', get_url)
        if get_url == True:# the == True is necessary so that strings will not false-positive
            play_fetch(self.set_cover_image, self.chipmunk.get_cover_image_url)
        else:
            play_fetch(self.set_cover_image_from_file, self.chipmunk.get_cover_image)

    def process_queue(self):
        if not self.chipmunk.process_queue(): return False
        self.clear_queue()
        return True

    def on_toolbutton_current_artist_clicked(self, widget):
        artist = self.chipmunk.current_song.artist
        station = self.chipmunk.get_station('artist', artist)
        self.play_station(station)

    def on_toolbutton_top_tag_clicked(self, widget):
        tag = self.chipmunk.current_song.tags[0]
        self.play_tag(tag)

    def play_tag(self, tag):
        station = self.chipmunk.get_station('tag', tag)
        self.play_station(station)

    def load_all_tabs(self):
        #print('loading tabs for', self.chipmunk.current_song.title, 'by', self.chipmunk.current_song.artist)
        self.load_lyrics_tab()
        self.load_artist_tab()
        self.load_track_tab()
        self.load_events_tab()

    def on_track_start(self, data):
        #TODO: Clean Up!
        if not self.process_queue():
            self.stopped = False
            station = self.chipmunk.current_station
            username = self.chipmunk.username
            title, artist, album, duration = data.split('[chipmunk-radio-separator]')
            duration = int(duration)
            song = Song(title, artist, album, duration)
            url = self.chipmunk.artist_url
            song.urls = {'artist': url, 'events': url+'/+events'}
            self.chipmunk.current_song = song
            self.set_cover_image(True)
            self.gui.set_track_labels(song)
            self.chipmunk.notify()
            label = self.gui['label_current_station']
            label.set_text(station.category.title()+': '+station.name)
            for widget in ['button_skip', 'button_share', 'button_tag',
            'button_ban', 'button_love', 'button_download', 'button_cover',
            'togglebutton_stop_delay', 'volumebutton', 'button_share',
            'togglebutton_scrobble', 'togglebutton_discovery']:
                self.gui[widget].set_sensitive(True)
            for widget in ['toolbutton_debookmark', 'toolbutton_bookmark']:
                self.gui[widget].hide()
            self.update_recent_toolbar()

            #widget = self.gui['toolbutton_my_library']
            widget = self.gui['toolbutton_my_mix']
            #if station != self.chipmunk.get_station('library', username):
            if station != self.chipmunk.get_station('mix', username):
                widget.show()
                self.add_recent(station)
                if station.category in ['artist', 'tag', 'group']:
                    if station in self.chipmunk.load('bookmarks'):
                        self.gui['toolbutton_bookmark'].hide()
                        self.gui['toolbutton_debookmark'].show()
                    else:
                        self.gui['toolbutton_debookmark'].hide()
                        self.gui['toolbutton_bookmark'].show()
            else:
                widget.hide()

            #add top tags for current track to the tag menu and toolbar
            play_fetch(self.update_tag_widgets, self.chipmunk.set_top_tags)
            #add current artist
            self.update_artists_menu()
            widget = self.gui['toolbutton_current_artist']
            if station != self.chipmunk.get_station('artist', artist):
                widget.set_label(artist)
                widget.show()
            else:
                widget.hide()
            time = str(duration // 60) + ':' + str(duration % 60).zfill(2)
            widget = self.gui['progressbar']
            widget.set_tooltip_text('Time Remaining, of ' + time)
            if self.chipmunk.current_station.rotate():
                self.chipmunk.queue_station(self.chipmunk.current_station)
            widget = self.gui['window_main']
            widget.set_title(' '.join([song.title, 'by', song.artist, '- Chipmunk']))
            self.load_all_tabs()
            #self.chipmunk.toggle_discovery_mode()

    def update_tag_widgets(self):
        self.update_tags_menu()
        self.update_top_tag_toolbutton()

    def on_toolbutton_station_clicked(self, widget, station):
        self.play_station(station)

    def update_recent_toolbar(self):
        toolbar = self.gui['toolbar_recent_stations']
        for toolbutton in toolbar:
            toolbar.remove(toolbutton)
        recent_stations = self.chipmunk.load('recent_stations')
        for station in recent_stations:
            blacklist = [self.chipmunk.current_station,
            self.chipmunk.get_station('artist', self.chipmunk.current_song.artist)]
            if station not in blacklist:
                toolbutton = self.gui.toolbutton_station(station)
                toolbar.insert(toolbutton, 0)

    def on_track_stop(self):
        self.gui.set_track_labels(Song(), False)
        self.gui['window_main'].set_title('Chipmunk')
        self.gui.clear_progress()
        self.clear_queue()
        for widget in ['button_download',
        'button_share',
        'button_tag',
        'button_ban',
        'button_love',
        'button_share',
        'button_skip',
        'togglebutton_stop_delay',
        'button_cover',
        'volumebutton']:
            self.gui[widget].set_sensitive(False)
        #self.gui['toolbutton_my_library'].show()
        self.gui['toolbutton_my_mix'].show()
        for widget in ['toolbutton_current_artist',
        'toolbutton_top_tag',
        'scrolledwindow_lyrics',
        'linkbutton_lyrics',
        'scrolledwindow_artist',
        'linkbutton_artist',
        'scrolledwindow_track',
        'linkbutton_track',
        'label_events_loading',
        'scrolledwindow_events',
        'linkbutton_events']:
            self.gui[widget].hide()
        for tab_id in ['lyrics', 'artist', 'track', 'events']:
            self.gui['label_'+tab_id+'_tab'].set_markup(tab_id.title())
        self.set_cover_image_from_file()
        self.stopped = True

    def on_button_cover_clicked(self, widget=None):
        if widget != None: bg(self.on_button_cover_clicked)
        elif self.chipmunk.current_song.album:
            webbrowser.open_new_tab(self.chipmunk.album_url)
        else: webbrowser.open_new_tab(self.chipmunk.track_url)

    def on_button_skip_clicked(self, widget=None):
        bg(self.chipmunk.skip_track)

    def on_button_love_clicked(self, widget):
        bg(self.chipmunk.love_track)
        widget.set_sensitive(False)

    def on_button_download_clicked(self, widget):
        #TODO
        #prevent the file from disappearing
        window = self.gui['filechooserdialog_download']
        #set the name
        window.show()

    def vbox_from_event_dict(self, event_dict):#GUI_GTK
        linkbutton = gtk.LinkButton(event_dict['url'], event_dict['title'])
        linkbutton.show()
        text = ''
        info_elements = ['venue', 'headliner', 'begins', 'description']
        for element in info_elements:
            info = event_dict[element]
            if info != '':
                if element in ['headliner']:
                    text += element.title()+': '+info+'\n'
                else:
                    text += info+'\n'
        text = text.strip()
        textview = gtk.TextView()
        try:
            textview.get_buffer().set_text(unicode(text))
        except:
            textview.get_buffer().set_text(text)
        textview.set_editable(False)
        textview.set_cursor_visible(False)
        textview.set_wrap_mode(gtk.WRAP_WORD)
        textview.show()
        #textview.grab_focus()#
        vbox = gtk.VBox()
        vbox.show()
        vbox.pack_start(linkbutton)
        vbox.pack_start(textview)
        #vbox.set_child_packing(textview, True, True, padding, pack_type)
        #textview.grab_focus()#
        return vbox

    def build_lyrics_tab(self, content):
        self.build_tab('lyrics', content)

    def build_artist_tab(self, content):
        self.build_tab('artist', content)

    def build_track_tab(self, content):
        self.build_tab('track', content)

    def build_events_tab(self, content):
        #print('Building Events tab ['+self.chipmunk.current_song.artist+']')
        self.build_tab('events', content)
        #print('Built ['+self.chipmunk.current_song.artist+']\n')

    def build_tab(self, tab_id, content):#This should probably move to GUI_GTK real soon
        if content == None:
            content = ''
        elif type(content) == type(''):
            content = content.strip()
        label = self.gui['label_'+tab_id+'_tab']
        label_text = filter_html_tags(label.get_label())
        scrolled_window = self.gui['scrolledwindow_'+tab_id]
        linkbutton = self.gui['linkbutton_'+tab_id]
        if tab_id == 'events':
            for child in scrolled_window.get_children():
                scrolled_window.remove(child)
            if len(content) == 0:
                scrolled_window.hide()
                linkbutton.show()
                label.set_markup_with_mnemonic(label_text)
            else:
                label.set_markup_with_mnemonic('<b>'+label_text+'</b>')
                linkbutton.hide()
                events_vbox = gtk.VBox(spacing=5)
                for event_dict in content:
                    event_vbox = self.vbox_from_event_dict(event_dict)
                    events_vbox.pack_start(event_vbox)
                scrolled_window.add_with_viewport(events_vbox)
                scrolled_window.show()
                events_vbox.show()
            self.gui['label_events_loading'].hide()
        else:
            textview = self.gui['textview_'+tab_id]
            if content == '' and self.chipmunk.current_song.title.strip() != '':
                linkbutton.show()
                scrolled_window.hide()
                label.set_markup_with_mnemonic(label_text)
            else:
                try:
                    textview.get_buffer().set_text(unicode(content, errors='replace'))
                except:
                    textview.get_buffer().set_text(content)
                scrolled_window.show()
                linkbutton.hide()
                label.set_markup_with_mnemonic('<b>'+label_text+'</b>')

    def load_lyrics_tab(self):
        play_fetch(self.build_lyrics_tab, self.chipmunk.get_lyrics)

    def load_artist_tab(self):
        play_fetch(self.build_artist_tab, self.chipmunk.get_artist_bio)

    def load_track_tab(self):
        play_fetch(self.build_track_tab, self.chipmunk.get_track_wiki)

    def load_events_tab(self):
        self.gui['linkbutton_events'].hide()
        self.gui['scrolledwindow_events'].hide()
        self.gui['label_events_loading'].show()
        play_fetch(self.build_events_tab, self.chipmunk.get_artist_events)

    def on_linkbutton_lyrics_clicked(self, widget, song=None):
        #watch the repetition, I should abstract some method...
        if song == None: song = self.chipmunk.current_song
        webbrowser.open_new_tab(song.lyrics_url)

    def on_linkbutton_youtube_clicked(self, widget):
        webbrowser.open_new_tab(self.song_to_share.youtube_url)

    def on_linkbutton_beemp3_clicked(self, widget):
        webbrowser.open_new_tab(self.song_to_share.beemp3_url)

    def on_linkbutton_lyrics_include_clicked(self, widget):
        if self.song_to_share.lyrics_url == None:
            self.on_linkbutton_lyrics_clicked(widget, self.song_to_share)
        else: webbrowser.open_new_tab(self.song_to_share.lyrics_url)

    def on_linkbutton_artist_include_clicked(self, widget):
        webbrowser.open_new_tab(self.song_to_share.artist_url)

    def on_linkbutton_events_include_clicked(self, widget):
        webbrowser.open_new_tab(self.song_to_share.events_url)

    def on_linkbutton_events_clicked(self, widget):
        webbrowser.open_new_tab(self.chipmunk.current_song.events_url)

    def on_linkbutton_track_clicked(self, widget):
        webbrowser.open_new_tab(self.chipmunk.track_url)

    def on_linkbutton_artist_clicked(self, widget):
        webbrowser.open_new_tab(self.chipmunk.artist_url)

    def on_button_share_clicked(self, widget):
        self.song_to_share = self.chipmunk.current_song.copy()
        self.gui.prepare_share_dialog()

    def on_volumebutton_value_changed(self, widget, vol=1):
        #relies on work done elsewhere (see update_progress_loop and __init__)
        self.target_volume = int(vol*100)

    def on_button_share_cancel_clicked(self, widget):
        self.gui['dialog_share'].hide()

    def on_button_share_send_clicked(self, widget):

        def share():
            entry = self.gui['comboboxentry_share'].child
            user_list = [user.strip() for user in entry.get_text().split(',')]

            buf = self.gui['textview_share_message'].get_buffer()
            message = buf.get_text(buf.get_start_iter(), buf.get_end_iter())

            for extra in ['youtube', 'beemp3', 'lyrics', 'artist', 'events']:
                if self.gui['checkbutton_'+extra].get_active():
                    url = self.song_to_share.get_url(extra)
                    try: url = surl.services.supportedServices()['chilp.it'].get([], url)
                    except: pass
                    message += '    \n'+extra.title()+': '+url

            self.chipmunk.share_track(user_list, message, self.song_to_share)

            if self.gui['checkbutton_share_shout'].get_active():
                self.chipmunk.save('shout_when_share', True)
                self.chipmunk.shout_track(user_list, message, self.song_to_share)
            else: self.chipmunk.save('shout_when_share', False)

            self.chipmunk.notify('Track Shared',
            '{} sent to {}'.format(self.song_to_share, ', '.join(user_list)),
            'emblem-shared')

        self.gui['dialog_share'].hide()
        bg(share)

    def on_button_tag_clicked(self, widget):
        #TODO
        pass

    def on_button_ban_clicked(self, widget): bg(self.chipmunk.ban_track)

    def on_button_clear_queue_clicked(self, widget): self.clear_queue()

    def on_toolbutton_bookmark_clicked(self, widget):
        station = self.chipmunk.current_station
        self.bookmark_station()

    def on_toolbutton_debookmark_clicked(self, widget):
        self.debookmark_station()

    def login(self, service_id='lastfm'):
        stored_username = self.chipmunk.load(service_id + '_username')
        stored_password = self.chipmunk.load(service_id + '_password')
        widget = self.gui['entry_login_username']
        widget.set_text(stored_username)
        widget.grab_focus()
        self.gui['entry_login_password'].set_text(stored_password)
        self.gui['dialog_login'].show()

    def on_menuitem_manage_lastfm_activate(self, widget):
        self.login('lastfm')

    def on_button_login_cancel_clicked(self, widget):
        self.gui['dialog_login'].hide()

    def on_button_login_ok_clicked(self, widget):
        username = self.gui['entry_login_username'].get_text()
        password = self.gui['entry_login_password'].get_text()
        if username and password: self.connect(username, password)
        self.gui['dialog_login'].hide()
        self.chipmunk.notify('Login and Password Set',
        'Click the play button or choose a station to begin', 'dialog-password')

    def build_friends_menu(self, users):
        menuitem_name = 'menuitem_friends'
        category = 'friend'
        parent = self.gui[menuitem_name]
        new_menu = gtk.Menu()
        for user in users:
            station = self.chipmunk.get_station('library', user, category)
            new_menu.append(self.gui.menuitem_station(station))
        parent.remove_submenu()
        parent.set_submenu(new_menu)

    def build_neighbors_menu(self, users):
        menuitem_name = 'menuitem_neighbors'
        category = 'neighbor'
        parent = self.gui[menuitem_name]
        new_menu = gtk.Menu()
        for user in users:
            station = self.chipmunk.get_station('library', user, category)
            new_menu.append(self.gui.menuitem_station(station))
        parent.remove_submenu()
        parent.set_submenu(new_menu)

    def update_artists_menu(self): self.build_artists_menu()

    def update_tags_menu(self, fetch=False):
        if fetch: tags = self.chipmunk.top_tags
        else: tags = self.chipmunk.current_song.tags
        self.build_tags_menu(tags)

    def update_groups_menu(self): self.build_groups_menu()

    def update_friends_menu(self):
        play_fetch(self.build_friends_menu, self.chipmunk.get_friends)

    def update_neighbors_menu(self):
        play_fetch(self.build_neighbors_menu, self.chipmunk.get_neighbors)

    def set_subscriber_settings(self, subscribes=None):
        if subscribes == None:
            play_fetch(self.set_subscriber_settings, self.chipmunk.is_subscriber)
        else:
            self.gui['menuitem_my_loved'].set_sensitive(subscribes)
            self.gui['menuitem_guest_loved'].set_sensitive(subscribes)
            self.chipmunk.save('is_subscriber', subscribes)

    def on_menuitem_manage_development_activate(self, widget):
        webbrowser.open_new_tab('http://bitbucket.org/EvilGnome/chipmunk-radio/issues?status=new&status=open')

    def on_menuitem_fit_toggled(self, widget):
        self.gui.fit_on_track_start = widget.get_active()

    def on_menuitem_notifications_toggled(self, widget):
        self.chipmunk.save('notifications_enabled', widget.get_active())

    def on_menuitem_somafm_website_activate(self, widget):
        webbrowser.open_new_tab('http://somafm.com/')

    def on_toolbutton_my_library_clicked(self, widget):
        station = self.chipmunk.get_station('library', self.chipmunk.username)
        self.play_station(station)

    def on_toolbutton_my_mix_clicked(self, widget):
        station = self.chipmunk.get_station('mix', self.chipmunk.username)
        self.play_station(station)

def print_help():
    print('chipmunk-radio [--play] [--next] [--stop-after-track] [--notify]',
          '[--love] [--ban] [--share] [--force-disconnect] [--help]')

if __name__ == '__main__':
    args = sys.argv[1:]
    if len(args) == 0:
        fp = open('/tmp/chipmunk-radio.pid', 'w')
        try: fcntl.lockf(fp, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError: sys.exit(0)# ^^ magic ensuring single instance
        app = ChipmunkApp()
        gtk.main()
    else:
        dbus_opts = {'--next': 'Next', '--play': 'Play',
                     '--stop-after-track': 'Stop', '--notify': 'Previous',
                     '--love': 'Love', '--ban': 'Ban', '--share': 'Share'}
        for x in args:
            try:
                subprocess.call(['dbus-send',
                '/org/bitbucket/chipmunk',
                'org.bitbucket.chipmunk.MediaPlayerKeyPressed',
                'string:'+dbus_opts[x]])
            except:
                if x == '--force-disconnect':
                    subprocess.call(['pkill', 'shell-fm'])
                    subprocess.call(['pkill', '-9', 'shell-fm'])
                else: print_help()
